
#include <stdio.h>
#include <stdlib.h>
#include "openbwts.h"
#include "time.h"

#ifndef __min
#define __min(x,y) (((x)<(y))?(x):(y))
#define __max(x,y) (((x)>(y))?(x):(y))
#endif

typedef unsigned short word;
typedef unsigned int   uint;
typedef unsigned char  byte;
typedef unsigned long long qword;
typedef   signed long long sqword;
typedef signed short sint16;

template <class T, int N> void bzero( T (&p)[N] ) { int i; for( i=0; i<N; i++ ) p[i]=0; }

uint flen( FILE* f ) {
  fseek( f, 0, SEEK_END );
  uint len = ftell(f);
  fseek( f, 0, SEEK_SET );
  return len;
}

#include "sh_qsort.inc"

byte* f_buf; uint f_len;
byte* g_buf;
uint* f_idx;
uint blocksize = 100<<20; // 100M


struct TEST {
  // Returns neg if 1<2, 0 if 1=2, pos if 1>2.
  static int  c( uint* A, int x, int y ) {
//    if( x==y ) return 0;
    uint px = A[x];
    uint py = A[y];
    int r,s;
    s = (px<py) ? SWAP(px,py),0x80000000 : 0;
    byte* p = &f_buf[px]; byte* th = &f_buf[blocksize];
    byte* q = &f_buf[py];
    for(; (r=1,p<th) && ((r=(*p)-(*q))==0); p++,q++ );
    return r^s;
  }
  static void s( uint* A, int x, int y ) {
    SWAP( A[x], A[y] );
  }
};

const int CNUM = 256;
uint freq[CNUM+1];

#include "postcoder.inc"

Postcoder PC;

//int    BWTS(byte* x1, byte*, int*, int);
//int    UnBWTS(byte*, byte*, int*, int);
double  t1,t2,t3;
double   xxx,yyy;

int main( int argc, char** argv ) {

  int i,j;
t1 = (double)clock();
t2 = t1;

  if( argc<4 ) return 1;
  FILE* f = fopen( argv[2], "rb" );
  if( f==0 ) return 2;
  FILE* g = fopen( argv[3], "wb" );
  if( g==0 ) return 3;

  if( argv[1][0]=='c' ) {

    j = atoi( &argv[1][1] );
    if( (j>=1) && (j<40000) ) blocksize=j*100000;
                       	
    f_len = flen(f);
    printf( "filelength = %i\n", f_len );
    blocksize = __min( blocksize, f_len );
    printf( "blocksize = %i\n", blocksize );

    f_idx = new uint[blocksize];
    if( f_idx==0 ) return 4;
    f_buf = new byte[blocksize];
    if( f_buf==0 ) return 5;

    PC.stor_init( g );
    PC.stor_uint( f_len );
    PC.stor_uint( blocksize );

    int blkI;
    for( blkI=0; f_len>0 ; blkI++, f_len-=blocksize ) {

t3 = (double)clock();
xxx = (t3 -t1)/CLOCKS_PER_SEC;
yyy = (t3 -t2)/CLOCKS_PER_SEC;
t2 = t3;
printf(" top of read block loop %f total %f \n", xxx,yyy);

      blocksize = __min( blocksize, f_len );
      fread( f_buf, 1,blocksize, f );
      
      {
       int i,j,k;
       for(i = 0,j = blocksize-1; i<j ;i++,j--){k = f_buf[i]; f_buf[i] = f_buf[j]; f_buf[j] = k;}
      }

      printf( "block #%i data loading complete (%i bytes).\n", blkI, blocksize );

/*
      for( i=0; i<blocksize; i++ ) f_idx[i]=i;

      printf( "index init complete.\n" );

      sh_qsort<TEST>( f_idx, 0, blocksize-1 );

      printf( "sort complete.\n" );

      int idx=0;
      for( i=0; i<blocksize; i++ ) {
        int j = f_idx[i];
        if( j==0 ) idx=i; else PC.stor_byte( f_buf[j-1] );
      }
      PC.stor_byte( f_buf[blocksize-1] );
*/
t3 = (double)clock();
xxx = (t3 -t1)/CLOCKS_PER_SEC;
yyy = (t3 -t2)/CLOCKS_PER_SEC;
t2 = t3;
printf(" start of BWTS block  %f total %f \n", xxx,yyy);

   
    int idx=0;
    BWTS(f_buf, f_buf, (int *)f_idx, (int)blocksize);
t3 = (double)clock();
xxx = (t3 -t1)/CLOCKS_PER_SEC;
yyy = (t3 -t2)/CLOCKS_PER_SEC;
t2 = t3;
printf(" start of compression block  %f total %f \n", xxx,yyy);

    for( i=0; i<blocksize; i++ ) PC.stor_byte(f_buf[i]); 

//      PC.stor_uint( idx );
      printf( "block BWT storing complete. index = %i\n", idx );
    }

  } else { // decoding

    PC.load_init(f);
    f_len = PC.load_uint();
    printf( "filelength = %i\n", f_len );
    blocksize = PC.load_uint();
    printf( "blocksize = %i\n", blocksize );

    f_idx = new uint[blocksize+1];
    if( f_idx==0 ) return 4;
    f_buf = new byte[blocksize];
    if( f_buf==0 ) return 5;

    int blkI;
    for( blkI=0; f_len>0 ; blkI++, f_len-=blocksize ) {
t3 = (double)clock();
xxx = (t3 -t1)/CLOCKS_PER_SEC;
yyy = (t3 -t2)/CLOCKS_PER_SEC;
t2 = t3;
printf(" top of loop doing decompression %f total %f \n", xxx,yyy);


      blocksize = __min( blocksize, f_len );

      for( i=0; i<blocksize; i++ ) f_buf[i]=PC.load_byte();

//      uint idx = PC.load_uint();

      printf( "block #%i data loading complete (%i bytes).\n", blkI, blocksize );

//      printf( "index = %i\n", idx );
t3 = (double)clock();
xxx = (t3 -t1)/CLOCKS_PER_SEC;
yyy = (t3 -t2)/CLOCKS_PER_SEC;
t2 = t3;
printf(" start of UnBWTS block  %f total %f \n", xxx,yyy);

      UnBWTSSR(f_buf, g, (int *)f_idx, (int)blocksize);
t3 = (double)clock();
xxx = (t3 -t1)/CLOCKS_PER_SEC;
yyy = (t3 -t2)/CLOCKS_PER_SEC;
t2 = t3;
printf(" start of write file block loop %f total %f \n", xxx,yyy);

//      for( i=0; i<blocksize; i++ ) putc( g_buf[i],g); 
/*
      bzero(freq);
      for( i=0; i<blocksize; i++ ) freq[f_buf[i]+1]++;
      for( i=0; i<CNUM; i++ ) freq[i+1]+=freq[i];

      f_idx[blocksize]=idx;
      for( i=0; i<blocksize; i++ ) {
        int c = f_buf[i];
        f_idx[freq[c]++]=i+(i>=idx);
      }

      j = idx;
      for( i=0; i<blocksize; i++ ) {
        j = f_idx[j];
        putc( f_buf[j-(j>=idx)], g );
      }
*/

      printf( "block unBWT storing complete.\n" );
    }

  } 

  PC.rc.Quit();

  fclose( g );
  fclose( f );
t3 = (double)clock();
xxx = (t3 -t1)/CLOCKS_PER_SEC;
yyy = (t3 -t2)/CLOCKS_PER_SEC;
t2 = t3;
printf(" end of program %f total %f \n", xxx,yyy);

 
  return 0;
}
